package com.qmth.themis.exam.websocket;//package com.qmth.themis.backend.websocket;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.qmth.themis.business.constant.SpringContextHolder;
import com.qmth.themis.business.constant.SystemConstant;
import com.qmth.themis.business.entity.TBSession;
import com.qmth.themis.business.util.RedisUtil;
import com.qmth.themis.common.enums.ExceptionResultEnum;
import com.qmth.themis.common.exception.BusinessException;
import com.qmth.themis.common.signature.SignatureInfo;
import com.qmth.themis.common.signature.SignatureType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Component;

import javax.websocket.*;
import javax.websocket.server.PathParam;
import javax.websocket.server.ServerEndpoint;
import java.io.IOException;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;

/**
 * @Description: websocker服务端
 * @Param:
 * @return:
 * @Author: wangliang
 * @Date: 2020/7/10
 */
@ServerEndpoint("/oe")
@Component
public class WebSocketServer
//        implements MessageListenerConcurrently
{
    private final static Logger log = LoggerFactory.getLogger(WebSocketServer.class);
    private volatile static ConcurrentHashMap<String, WebSocketServer> webSocketMap = new ConcurrentHashMap<>();
    /**
     * 与某个客户端的连接会话，需要通过它来给客户端发送数据
     */
    private Session session;
    private String sessionId;
    private String platform = null;
    private String deviceId = null;
    private String Authorization = null;
    private Long time = null;
    private RedisUtil redisUtil;

    /**
     * 连接建立成功调用的方法
     */
    @OnOpen
    public void onOpen(Session session) {
        Map<String, List<String>> mapParameter = session.getRequestParameterMap();
        if (Objects.isNull(mapParameter)) {
            throw new BusinessException(ExceptionResultEnum.PARAMS_ILLEGALITY);
        }
        if (Objects.isNull(mapParameter.get("platform")) || mapParameter.get("platform").size() == 0) {
            throw new BusinessException(ExceptionResultEnum.PLATFORM_INVALID);
        }
        this.platform = String.valueOf(mapParameter.get("platform").get(0));
        if (Objects.isNull(mapParameter.get("deviceId")) || mapParameter.get("deviceId").size() == 0) {
            throw new BusinessException(ExceptionResultEnum.DEVICE_ID_INVALID);
        }
        this.deviceId = String.valueOf(mapParameter.get("deviceId").get(0));
        this.Authorization = String.valueOf(mapParameter.get("Authorization").get(0));
        this.time = Long.parseLong(String.valueOf(mapParameter.get("time").get(0)));
        String method = SystemConstant.GET;
        final SignatureInfo info = SignatureInfo
                .parse(Authorization);
        if (Objects.nonNull(info) && info.getType() == SignatureType.TOKEN) {
            String sessionId = info.getInvoker();
            redisUtil = SpringContextHolder.getBean(RedisUtil.class);
            TBSession tbSession = (TBSession) redisUtil.getUserSession(sessionId);
            if (Objects.isNull(tbSession)) {
                throw new BusinessException(ExceptionResultEnum.LOGIN_NO);
            } else {
                if (info.validate(tbSession.getAccessToken()) && info.getTimestamp() < tbSession.getExpireTime().getTime()
                        && platform.equalsIgnoreCase(tbSession.getPlatform()) && Objects.equals(deviceId, tbSession.getDeviceId())) {
                    this.session = session;
                    session.setMaxIdleTimeout(SystemConstant.WEBSOCKET_MAX_TIME_OUT);
                    this.sessionId = tbSession.getId();
                    if (webSocketMap.containsKey(session)) {
                        webSocketMap.remove(sessionId);
                        webSocketMap.put(sessionId, this);
                        //加入set中
                    } else {
                        webSocketMap.put(sessionId, this);
                        //加入set中
                        addOnlineCount();
                        //在线数加1
                    }
                    log.info("用户连接:" + this.sessionId + ",当前在线人数为:" + getOnlineCount());
                    try {
                        sendMessage("连接成功");
                    } catch (IOException e) {
                        e.printStackTrace();
                        log.error("用户:" + this.sessionId + ",网络异常!!!!!!");
                    }
                } else {
                    throw new BusinessException(ExceptionResultEnum.AUTHORIZATION_ERROR);
                }
            }
        } else {
            throw new BusinessException(ExceptionResultEnum.AUTHORIZATION_ERROR);
        }
    }

    /**
     * 连接关闭调用的方法
     */
    @OnClose
    public void onClose() {
        if (webSocketMap.containsKey(this.sessionId)) {
            webSocketMap.remove(this.sessionId);
            //从set中删除
            subOnlineCount();
        }
        log.info("用户退出:" + this.sessionId + ",当前在线人数为:" + getOnlineCount());
    }

    /**
     * 收到客户端消息后调用的方法
     *
     * @param message
     * @param session
     */
    @OnMessage
    public void onMessage(String message, Session session) {
        //可以群发消息
        //消息保存到数据库、redis
        if (Objects.nonNull(message)) {
            try {
                //解析发送的报文
                JSONObject jsonObject = JSON.parseObject(message);
                //追加发送人(防止串改)
                jsonObject.put("fromSessionId", this.sessionId);
                String toSessionId = jsonObject.getString("toSessionId");
                //传送给对应toSessionId用户的websocket
                if (Objects.nonNull(toSessionId) && webSocketMap.containsKey(toSessionId)) {
                    webSocketMap.get(toSessionId).sendMessage(jsonObject.toJSONString());
                    log.info("发送消息jsonObject:{}", jsonObject.toJSONString());
                } else {
                    log.error("请求的sessionId:" + toSessionId + "不在该服务器上");
                    //否则不在这个服务器上，发送到mysql或者redis
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * 错误
     *
     * @param session
     * @param error
     */
    @OnError
    public void onError(Session session, Throwable error) {
        log.error("用户错误:" + this.sessionId + ",原因:" + error.getMessage());
        error.printStackTrace();
        throw new BusinessException(error.getMessage());
    }

    /**
     * 实现服务器主动推送
     *
     * @param message
     * @throws IOException
     */
    public void sendMessage(String message) throws IOException {
        log.info("message:{}", message);
    }

    /**
     * 发送自定义消息
     */
    public static void sendInfo(String message, @PathParam("sessionId") String sessionId) throws IOException {
        log.info("发送消息到:" + sessionId + "，报文:" + message);
        if (Objects.nonNull(sessionId) && webSocketMap.containsKey(sessionId)) {
            webSocketMap.get(sessionId).sendMessage(message);
        } else {
            log.error("用户" + sessionId + ",不在线！");
        }
    }

    /**
     * 获取在线人数
     *
     * @return
     */
    public synchronized int getOnlineCount() {
        Object o = redisUtil.getCacheValue(SystemConstant.WEBSOCKET_OE_ONLINE_COUNT, "");
        return Objects.isNull(o) ? 0 : (int) o;
    }

    /**
     * 在线人数加一
     */
    public synchronized void addOnlineCount() {
        if (redisUtil.lock(SystemConstant.REDIS_LOCK_WEBSOCKET_PREFIX, this.sessionId, SystemConstant.REDIS_LOCK_WEBSOCKET_TIME_OUT, TimeUnit.SECONDS)) {
            Object o = redisUtil.getCacheValue(SystemConstant.WEBSOCKET_OE_ONLINE_COUNT, "");
            int count = 0;
            if (Objects.nonNull(o)) {
                count = (int) o;
            }
            count++;
            redisUtil.setCacheValue(SystemConstant.WEBSOCKET_OE_ONLINE_COUNT, "", count);
        }
    }

    /**
     * 在线人数减一
     */
    public synchronized void subOnlineCount() {
        if (redisUtil.lock(SystemConstant.REDIS_LOCK_WEBSOCKET_PREFIX, this.sessionId, SystemConstant.REDIS_LOCK_WEBSOCKET_TIME_OUT, TimeUnit.SECONDS)) {
            Object o = redisUtil.getCacheValue(SystemConstant.WEBSOCKET_OE_ONLINE_COUNT, "");
            int count = 0;
            if (Objects.nonNull(o)) {
                count = (int) o;
            }
            count--;
            redisUtil.setCacheValue(SystemConstant.WEBSOCKET_OE_ONLINE_COUNT, "", count < 0 ? 0 : count);
        }
    }

//    @Override
//    public ConsumeConcurrentlyStatus consumeMessage(List<MessageExt> msgs, ConsumeConcurrentlyContext consumeConcurrentlyContext) {
//        try {
//            long threadId = Thread.currentThread().getId();
//            String threadName = Thread.currentThread().getName();
//            for (MessageExt messageExt : msgs) {
//                log.info(":{}-:{} websocketConsumer 重试次数:{}", threadId, threadName, messageExt.getReconsumeTimes());
//                MqDto mqDto = JacksonUtil.readJson(new String(messageExt.getBody(), Constants.CHARSET), MqDto.class);
//                log.info(":{}-:{} websocketConsumer 接收到的消息:{}", threadId, threadName, JacksonUtil.parseJson(mqDto));
//                log.info(":{}-:{} websocketConsumer mqDto sequence:{},tag:{}", threadId, threadName, mqDto.getSequence(), mqDto.getTag());
//                Map map = mqDto.getProperties();
//                String body = JacksonUtil.parseJson(mqDto.getBody());
//                log.info("map:{},body:{}", JacksonUtil.parseJson(map), body);
//                String model = String.valueOf(map.get("model"));
//                MessageModel messageModel = MessageModel.valueOf(model);
//                if (messageModel.ordinal() == MessageModel.CLUSTERING.ordinal()) {
//                    webSocketMap.get(map.get("toUserId")).sendMessage(body);
//                } else {
//                    webSocketMap.forEach((k, v) -> {
//                        try {
//                            v.sendMessage(body);
//                        } catch (IOException e) {
//                            e.printStackTrace();
//                        }
//                    });
//                }
//            }
//        } catch (Exception e) {
//            e.printStackTrace();
//            return ConsumeConcurrentlyStatus.RECONSUME_LATER;//重试
//        }
//        return ConsumeConcurrentlyStatus.CONSUME_SUCCESS;//成功
//    }
//
//    @Service
//    @RocketMQMessageListener(consumerGroup = "websocketConsumerImGroup", topic = "websocketImTopic", selectorType = SelectorType.TAG, selectorExpression = "*")
//    public class sessionConsumerWeb implements RocketMQListener<Message>, RocketMQPushConsumerLifecycleListener {
//
//        @Override
//        public void onMessage(Message message) {
//            //实现RocketMQPushConsumerLifecycleListener监听器之后，此方法不调用
//        }
//
//        @Override
//        public void prepareStart(DefaultMQPushConsumer defaultMQPushConsumer) {
//            defaultMQPushConsumer.setConsumeMessageBatchMaxSize(SystemConstant.CONSUME_MESSAGE_BATCH_MAX_SIZE);//每次拉取10条
//            defaultMQPushConsumer.setConsumeFromWhere(ConsumeFromWhere.CONSUME_FROM_FIRST_OFFSET);
//            defaultMQPushConsumer.setMaxReconsumeTimes(SystemConstant.MAXRECONSUMETIMES);//最大重试次数
////            defaultMQPushConsumer.setMessageModel(MessageModel.BROADCASTING);
//            defaultMQPushConsumer.registerMessageListener(WebSocketServer.this::consumeMessage);
//        }
//    }
}

